


/*
 *	threshold.c - handle thresholding functions
 */



/*		CRYSTAL IMAGE 0.9.1 BETA
		This program is copyrighted material.
		(C) 2002-2010 by Mark A. Haidekker, all rights reserved.
		
		This program is free software: you can redistribute it and/or modify
		it under the terms of the GNU General Public License as published by
		the Free Software Foundation, either version 3 of the License, or
		(at your option) any later version.

		This program is distributed in the hope that it will be useful,
		but WITHOUT ANY WARRANTY; without even the implied warranty of
		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE:

		THERE IS NO WARRANTY FOR THE PROGRAM, TO THE EXTENT PERMITTED BY
		APPLICABLE LAW.  EXCEPT WHEN OTHERWISE STATED IN WRITING THE COPYRIGHT
		HOLDERS AND/OR OTHER PARTIES PROVIDE THE PROGRAM "AS IS" WITHOUT WARRANTY
		OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO,
		THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
		PURPOSE.  THE ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE PROGRAM
		IS WITH YOU.  SHOULD THE PROGRAM PROVE DEFECTIVE, YOU ASSUME THE COST OF
		ALL NECESSARY SERVICING, REPAIR OR CORRECTION.

		IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING
		WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MODIFIES AND/OR CONVEYS
		THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU FOR DAMAGES, INCLUDING ANY
		GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE
		USE OR INABILITY TO USE THE PROGRAM (INCLUDING BUT NOT LIMITED TO LOSS OF
		DATA OR DATA BEING RENDERED INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD
		PARTIES OR A FAILURE OF THE PROGRAM TO OPERATE WITH ANY OTHER PROGRAMS),
		EVEN IF SUCH HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF
		SUCH DAMAGES.

		For the full text of the license, please see the file COPYING
		in the project's main directory, or read it on the Web at
		http://www.gnu.org/copyleft/gpl.html

 */



#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <stdio.h>
#include <string.h>
#include <math.h>

#include <time.h>
#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>


#include "application.h"
#include "defs.h"
#include "dlgbox.h"
#include "functions.h"
#include "functions2.h"
#include "graphdsp.h"
#include "gtkhelpers.h"
#include "imgbuf.h"
#include "interface.h"
#include "miscop.h"
#include "roi.h"
#include "secondwnd.h"
#include "threshold.h"


/*******************************************************

	Threshold. This is called in a complicated way due
	to restrictions in gtk. It has to be installed as 
	a callback function to the threshold selection dialog.
	
	There are multiple methods: 
	
	TH_BINARY converts the image into a mask
	that contains only 0 and 1.
	
	TH_HARD sets all pixels to zero that are below the
	threshold, leaves the others untouched.
	
	TH_SOFT sets all pixels to zero that are below the
	threshold, and subtracts the threshold from all others,
	effectively reducing the image values by the threshold
	value, but clamping the values at a minimum. 
	
	TH_HARDSYM is the origin-symmetrical version of TH_HARD
	and clamps all values to zero with an ABSOLUTE VALUE 
	smaller than the threshold
	
	TH_SOFTSYM is the origin-symmetrical version of TH_SOFT
	
	TH_SATURATE is the opposite of thresholding, it clamps
	values above the threshold
	
	TH_SATSYM is a sign-symmetric saturation function
	
********************************************************/


/*
	TODO: Some functions don't observe the roi although they should
*/


int thresh_int (int origval, int threshold, THRESHOLD_TYPE ty)
{
int tval;

	if (ty==TH_BINARY)
	{
		if (origval >= threshold) tval=1; else tval=0;
	}
	else if (ty==TH_SOFT)
	{
		if (origval >= threshold) tval = origval-threshold; else tval=0;
	}
	else if (ty==TH_HARD)
	{
		if (origval >= threshold) tval = origval; else tval=0;
	}
	else if (ty==TH_SOFTSYM)
	{
		if (origval<0)				/* Must fully differentiate between sign cases */
		{
			if (origval < -abs(threshold)) 
				tval = 0;
			else
				tval = origval+abs(threshold);	/* negative origval reduced by threshold */
		}
		else						/* Values > 0 */
		{
			if (origval < abs(threshold)) 
				tval = 0;
			else
				tval = origval-abs(threshold);	/* positive origval reduced by threshold */
		}
	}
	else if (ty==TH_HARDSYM)
	{
		if (origval < 0)
		{
			if (-origval <= -abs(threshold)) tval = origval; else tval=0;
		}
		else
		{
			if (origval >= abs(threshold)) tval = origval; else tval=0;
		}
	}
	else if (ty==TH_SATURATE)
	{
		if (origval >= threshold) tval = threshold; else tval=origval;
	}
	else if (ty==TH_SATSYM)
	{
		if (origval < 0)
		{
			if (-origval <= -abs(threshold)) tval = -abs(threshold); else tval=origval;
		}
		else
		{
			if (origval >= abs(threshold)) tval = abs(threshold); else tval=origval;
		}
	}
	else
		tval = origval;
		
	return tval;
}


float thresh_flt (float origval, float threshold, THRESHOLD_TYPE ty)
{
float tval;

	if (ty==TH_BINARY)
	{
		if (origval >= threshold) tval=1.0; else tval=0.0;
	}
	else if (ty==TH_SOFT)
	{
		if (origval >= threshold) tval = origval-threshold; else tval=0.0;
	}
	else if (ty==TH_HARD)
	{
		if (origval >= threshold) tval = origval; else tval=0.0;
	}
	else if (ty==TH_SOFTSYM)
	{
		if (origval<0)				/* Must fully differentiate between sign cases */
		{
			if (origval < -fabs(threshold)) 
				tval = 0;
			else
				tval = origval+fabs(threshold);	/* negative origval reduced by threshold */
		}
		else						/* Values > 0 */
		{
			if (origval < fabs(threshold)) 
				tval = 0;
			else
				tval = origval-fabs(threshold);	/* positive origval reduced by threshold */
		}
	}
	else if (ty==TH_HARDSYM)
	{
		if (origval < 0)
		{
			if (-origval <= -fabs(threshold)) tval = origval; else tval=0;
		}
		else
		{
			if (origval >= fabs(threshold)) tval = origval; else tval=0;
		}
	}
	else if (ty==TH_SATURATE)
	{
		if (origval >= threshold) tval = threshold; else tval=origval;
	}
	else if (ty==TH_SATSYM)
	{
		if (origval < 0)
		{
			if (-origval <= -fabs(threshold)) tval = -fabs(threshold); else tval=origval;
		}
		else
		{
			if (origval >= fabs(threshold)) tval = fabs(threshold); else tval=origval;
		}
	}
	else
		tval = origval;

	return tval;
}




/* This function gets indirectly called by the OK button handler of the
   threshold selection dialog box. We have all the information in the
   thresh_return structure, including a pointer to the image to work on */

void do_thresholding ()
{
int bx1, by1, bx2, by2;
int x,y,z;
char is_mainimg;
long i, idx, zidx;

	if (thresh_return.img == &mainimg) undo_nexus (thresh_return.img);

	if (thresh_return.img == &mainimg)
	{
		roi_bbox (thresh_return.img, &bx1, &by1, &bx2, &by2);	/* Observe ROI only on main image */
		is_mainimg=1;
	}
	else
	{
		bx1=0; by1=0; bx2=thresh_return.img->xmax-1; by2=thresh_return.img->ymax-1;
		is_mainimg=0;
	}

	dp (1, "Thresholding");
	if (is_mainimg) dp (1, " on main image");
	dp (1, " with threshold value ");
	if (thresh_return.img->imgtype==B_RGB)
		dp (1, " RGB (%d,%d,%d)",thresh_return.r,thresh_return.g,thresh_return.b);
	else
		dp (1, " %f",thresh_return.t);
	dp (1, " Function: ");
	if (thresh_return.thresh_type==TH_BINARY) dp (1, "Binary");
	else if (thresh_return.thresh_type==TH_SOFT) dp (1, "Soft");
	else if (thresh_return.thresh_type==TH_HARD) dp (1, "Hard");
	else if (thresh_return.thresh_type==TH_SOFTSYM) dp (1, "Soft - sign symmetric");
	else if (thresh_return.thresh_type==TH_HARDSYM) dp (1, "Hard - sign symmetric");
	else if (thresh_return.thresh_type==TH_SATURATE) dp (1, "Saturation function");
	else if (thresh_return.thresh_type==TH_SATSYM) dp (1, "Saturation - sign symmetric");
	dp (1, "\n");

	for (y=by1; y<=by2; y++)
	{
		idx = y*thresh_return.img->xmax;
		for (x=bx1; x<=bx2; x++)
		{
			if (!is_mainimg || in_roi(x,y))
			{
				switch (thresh_return.img->imgtype)
				{
					case B_CHAR:
					{
						unsigned char* p = thresh_return.img->data;
						for (z=0; z<thresh_return.img->zmax; z++)
						{
							zidx = z*thresh_return.img->xmax * thresh_return.img->ymax;
							p[idx+zidx+x] = thresh_int (p[idx+zidx+x],(int)thresh_return.t,thresh_return.thresh_type);
						}
						break;
					}
					case B_SHORT:
					{
						short* p = thresh_return.img->data;
						for (z=0; z<thresh_return.img->zmax; z++)
						{
							zidx = z*thresh_return.img->xmax * thresh_return.img->ymax;
							p[idx+zidx+x] = thresh_int (p[idx+zidx+x],(int)thresh_return.t,thresh_return.thresh_type);
						}
						break;
					}
					case B_FLOAT:
					{
						float* p = thresh_return.img->data;
						for (z=0; z<thresh_return.img->zmax; z++)
						{
							zidx = z*thresh_return.img->xmax * thresh_return.img->ymax;
							p[idx+zidx+x] = thresh_flt (p[idx+zidx+x],thresh_return.t,thresh_return.thresh_type);
						}
						break;
					}					
					case B_RGB:
					{
						unsigned char* p = thresh_return.img->data;
						for (z=0; z<thresh_return.img->zmax; z++)
						{
							zidx = z*thresh_return.img->xmax * thresh_return.img->ymax;
							i = 3*(idx+zidx+x);
							p[i] = thresh_int (p[i],thresh_return.r,thresh_return.thresh_type);
							p[i+1] = thresh_int (p[i+1],thresh_return.g,thresh_return.thresh_type);
							p[i+2] = thresh_int (p[i+2],thresh_return.b,thresh_return.thresh_type);
						}
						break;
					}
				}
			}
		}
	}
					
	make_minmax (thresh_return.img);
	thresh_return.img->uid = next_uid++;
	if (is_mainimg) 
	{
		make_imlib_buffer_from_main_image (cur_slice);
		redraw (1);
	}

}


/* Bi-level thresholding -- done only on one slice if slice>=0 or on all slices if slice==-1 */

void do_thresholding2 (float l, float u, int slice)
{
int bx1, by1, bx2, by2;
int x,y,z, stslice, enslice;
char is_mainimg;
long i, idx, zidx;
float val;

	if (thresh_return.img == &mainimg) undo_nexus (thresh_return.img);

	if (thresh_return.img == &mainimg)
	{
		roi_bbox (thresh_return.img, &bx1, &by1, &bx2, &by2);	/* Observe ROI on main image */
		is_mainimg=1;
	}
	else
	{
		bx1=0; by1=0; bx2=thresh_return.img->xmax-1; by2=thresh_return.img->ymax-1;
		is_mainimg=0;
	}

	if (slice<0) 
	{
		stslice=0; enslice=thresh_return.img->zmax-1;
	}
	else
	{
		stslice=slice; enslice=slice;
	}

	for (z=stslice; z<=enslice; z++)
	{
		for (y=by1; y<=by2; y++)
		{
			idx = y*thresh_return.img->xmax;
			for (x=bx1; x<=bx2; x++)
			{
				if (!is_mainimg || in_roi(x,y))
				{
					switch (thresh_return.img->imgtype)
					{
						case B_CHAR:
						{
							unsigned char* p = thresh_return.img->data;
							zidx = z*thresh_return.img->xmax * thresh_return.img->ymax;
							val = p[idx+zidx+x];
							if ((val>=l) && (val<=u)) p[idx+zidx+x] = 1; else p[idx+zidx+x] = 0;
							break;
						}
						case B_SHORT:
						{
							short* p = thresh_return.img->data;
							zidx = z*thresh_return.img->xmax * thresh_return.img->ymax;
							val = p[idx+zidx+x];
							if ((val>=l) && (val<=u)) p[idx+zidx+x] = 1; else p[idx+zidx+x] = 0;
							break;
						}
						case B_FLOAT:
						{
							float* p = thresh_return.img->data;
							zidx = z*thresh_return.img->xmax * thresh_return.img->ymax;
							val = p[idx+zidx+x];
							if ((val>=l) && (val<=u)) p[idx+zidx+x] = 1.0; else p[idx+zidx+x] = 0.0;
							break;
						}					
						case B_RGB:
						{
							unsigned char* p = thresh_return.img->data;
							zidx = z*thresh_return.img->xmax * thresh_return.img->ymax;
							i = 3*(idx+zidx+x);
							if ((p[i]>=l) && (p[i]<=u)) p[i] = 1; else p[i] = 0;
							i++;
							if ((p[i]>=l) && (p[i]<=u)) p[i] = 1; else p[i] = 0;
							i++;
							if ((p[i]>=l) && (p[i]<=u)) p[i] = 1; else p[i] = 0;
							break;
						}
					}
				}
			}
		}
	}
	make_minmax (thresh_return.img);
	thresh_return.img->uid = next_uid++;
	if (is_mainimg) 
	{
		make_imlib_buffer_from_main_image (cur_slice);
		redraw (1);
	}

}



/* GUI: Open the threshold dialogbox and obtain the threshold value(s).
   When OK is pressed, call the function supplied as third parameter.
   The image pointer will be stored in the thresh_return structure
   and thus passed to the actual threshold handler above */

void threshold_segment (image_type* img)
{

	select_threshold (img, 1, (gpointer)do_thresholding, "Threshold Image");
	
}



/******************************************************

	Iterative thresholding and Otsu's Method
	- Works on images with bimodal histograms. 
	
	Iterative:
	Selects some arbitraty initial
	threshold and computes the means values of the
	separated regions. Then computes a new threshold
	from the average of the two means. Repeat until
	the threshold no longer changes.
	
	Otsu's:
	Create a histogram and compute ratio of
	interclass variance to within-class variance.
	Test all gray levels to maximize this ratio.

	Entropy:
	Create a histogram and maximize the combined
	entropy of the two classes ny an exhaustive
	search of all threshold values

	First Valley:
	Create a smoothed histogram and search, starting
	from black, for the first valley, i.e., where the
	(hopefully decaying) slope drops below a threshold

*******************************************************/


/* Many auto-threshold functions operate on a histogram. The internal
	histogram size INT_HISTSIZE determines the number of bins used */

#define INT_HISTSIZE		256



/* part 1 (a) - determine the "optimum" threshold via isodata */

float threshold_iterative (image_type* img, int zslice)
{
int x,y;
float thresh,newthresh,buf,mean1,mean2, delta;
int cnt1,cnt2;
int abort, pass;

	abort=0; pass=0;
	
	/* Initial threshold collects everything in mean1. The effect of this
		"strange" initialization is that after one pass the threshold is
		the image mean. This, in turn, is a reasonable initialization. */
	
	if (img->imgtype==B_FLOAT)
	{
		thresh = img->minval.fmin-1.0;
		delta = img->maxval.fmax-img->minval.fmin;
	}
	else
	{
		thresh = img->minval.imin-1;
		delta = img->maxval.imax-img->minval.imin;
	}
	
	/* Start iterations */
	
	do
	{
		cnt1=0; cnt2=0; mean1=0.0; mean2=0.0;
		for (y=0; y<img->ymax; y++)
			for (x=0; x<img->xmax; x++)
			{
				buf = readbuf_flt (img,x,y,zslice);
				if (buf > thresh)
				{
					mean1 += buf; cnt1++;
				}
				else
				{
					mean2 += buf; cnt2++;
				}
			}
			
			if (cnt1>0) mean1 /= cnt1;
			if (cnt2>0) mean2 /= cnt2;
			
			if ((cnt1==0) && (cnt2==0))
				{ abort=1; newthresh=0; }		/* bad error */
			else if (cnt1==0)
				newthresh = mean2;
			else if (cnt2==0)
				newthresh = mean1;
			else
				newthresh = 0.5*(mean1+mean2);
			
		if ( fabs(thresh - newthresh) < 0.001*delta )
			abort=1;
		pass++;
		thresh = newthresh;
		dp (2, "Pass %d: Mean 1 %f (area %d), mean 2 %f (area %d), new threshold %f\n",
			pass, mean1, cnt1, mean2, cnt2, newthresh);
	}
	while (!abort);
	
	return thresh;
}


/* Part 1 (b): Determine threshold using Otsu's Method */


float otsus_threshold (image_type* img, int zslice, int mkgraph)
{
int x,y,i;
float min,max,thresh,buf,mean1,a,b,c,cmax, sig_max;
int cnt1;
int bin;
float histx[256], histy[256],histz[256], omega[256], mu[256];


	imgminmax (img, &min, &max);

	memset (histx, 0, sizeof (histx));
	memset (histy, 0, sizeof (histy));

	/* for graphing purposes, compute histx */
	
	for (i=0; i<INT_HISTSIZE; i++) histx[i] = min + i*(max-min)/INT_HISTSIZE;
	
	/* Generate the histogram with INT_HISTSIZE classes and compute
	   the image's mean gray level mean1 */
	
	cnt1=0; mean1=0.0;
	for (y=0; y<img->ymax; y++)
		for (x=0; x<img->xmax; x++)
		{
			buf = readbuf_flt (img,x,y,zslice);
			bin = (int) (INT_HISTSIZE*(buf-min)/(max-min));
			if (bin>INT_HISTSIZE-1) bin=INT_HISTSIZE-1; else if (bin<0) bin=0;
			histy[bin] += 1.0;
			mean1+=buf; cnt1++;
		}
	mean1 /= cnt1;
	
	/* The histogram must be area-normalized so that
	   sum from i=0 to 255 of histy[i] is 1.0.
	   (convert to probability density function) */
	   
	for (i=0; i<INT_HISTSIZE; i++) histy[i] /= cnt1;

	/* try all bins for the maximum. Can't use the last class for a divby0 */

	if (NULL)		/* Algorithm according to Petrou-book. Junk. */
	{
		a=0.0; b=0.0; cmax=0.0; thresh=min;
		for (i=0; i<255; i++)
		{
			a += histy[i];		/* theta(t) in Petrou-book */
			b += histx[i]*histy[i];	/* mu(t) in Petrou book */
			c = SQR(b*(mean1-b)) / (a*(1.0-a));
			dp (2, "t=%f   theta(t) = %f    mu(t) = %f   sb(t)= %f\n",histx[i],a,b,sqrt(c));
			histz[i]=c;
			if (c>cmax) { cmax=c; thresh=histx[i]; }
		}
	}
	else		/* Another algorithm realization */
	{
		/* Iteratively compute omega and mu */
		
		omega[0]=histy[0]; mu[0]=0;
		for (i=1; i<INT_HISTSIZE; i++)
		{
			omega[i] = omega[i-1] + histy[i];	/* Integrated prob density */
			mu[i] = mu[i-1] + i*histy[i];
		}
		
		/* Maximization of inter-class variance sigma (stored in histz) */

		thresh = 0;
		sig_max = 0.0;
		for (i = 0; i < INT_HISTSIZE-1; i++) 
		{
			if ( (omega[i]!=0.0) && (omega[i] != 1.0) )
				histz[i] = 	pow(mu[INT_HISTSIZE-1]*omega[i] - mu[i], 2) / 
							(omega[i]*(1.0 - omega[i]));
			else
				histz[i] = 0.0;

			if (histz[i] > sig_max) 
			{
				sig_max = histz[i];
				thresh = histx[i];
			}
		}


	}
	
	if (mkgraph && !macrorun) graphme (histx,histz,INT_HISTSIZE);
	dp (1, "Otsu's threshold: %f\n", thresh);
	return thresh;

}



/* Part 1 (c) - entropy-maximizing threshold */

float entropy_threshold (image_type* img, int zslice, int mkgraph)
{
int x,y,i,j;
float min,max,thresh,buf,h1,h2,ent_max, cumul_t;
int cnt1;
int bin;
float histx[INT_HISTSIZE], histy[INT_HISTSIZE],histz[INT_HISTSIZE], omega[INT_HISTSIZE], mu[INT_HISTSIZE];


	imgminmax (img, &min, &max);

	memset (histx, 0, sizeof (histx));
	memset (histy, 0, sizeof (histy));

	/* for graphing purposes, compute histx */
	
	for (i=0; i<INT_HISTSIZE; i++) histx[i] = min + i*(max-min)/INT_HISTSIZE;
	
	/* Generate the histogram with INT_HISTSIZE classes */
	
	cnt1=0;
	for (y=0; y<img->ymax; y++)
		for (x=0; x<img->xmax; x++)
		{
			buf = readbuf_flt (img,x,y,zslice);
			bin = (int) (INT_HISTSIZE*(buf-min)/(max-min));
			if (bin>INT_HISTSIZE-1) bin=INT_HISTSIZE-1; else if (bin<0) bin=0;
			histy[bin] += 1.0;
			cnt1++;
		}
	
	/* The histogram must be area-normalized so that
	   sum from i=0 to 255 of histy[i] is 1.0.
	   (convert to probability density function) */
	   
	for (i=0; i<INT_HISTSIZE; i++) histy[i] /= cnt1;

	/* try all bins for the maximum. Can't use the last class for a divby0 */
	

	thresh = 0;
	ent_max = 0.0; cumul_t=0.0;
	for (i = 0; i < INT_HISTSIZE-1; i++) 		/* Try all bins */
	{
		cumul_t += histy[i];				/* Cumulative probability from 0 to i */
		
		h1=0.0; h2=0.0;
		for (j=0; j<i; j++)     if (cumul_t>0.0)  h1 += (histy[j]*SAFELOG2(histy[j]/cumul_t)/cumul_t);
		for (j=i; j<INT_HISTSIZE-1; j++) if (cumul_t!=1.0) h2 += (histy[j]*SAFELOG2(histy[j]/(1-cumul_t))/(1-cumul_t));
		histz[i] = -h1-h2;
		
		if (histz[i] > ent_max) 
		{
			ent_max = histz[i];
			thresh = histx[i];
		}
	}

	if (mkgraph && !macrorun) graphme (histx,histz,INT_HISTSIZE);
	dp (1, "Entropy threshold: %f\n", thresh);
	return thresh;

}


/* Part 1 (d): Determine threshold by finding the first valley */

float fvalley_threshold (image_type* img, int zslice, int mkgraph)
{
int x,y,i,j;
float min,max,thresh,buf,mean1,a,b,c,cmax, peak;
int cnt1, status;
int bin;
float histx[INT_HISTSIZE], histy[INT_HISTSIZE],histz[3*INT_HISTSIZE], omega[INT_HISTSIZE], mu[INT_HISTSIZE];


	imgminmax (img, &min, &max);

	memset (histx, 0, sizeof (histx));
	memset (histy, 0, sizeof (histy));

	/* Generate the normalized histogram with INT_HISTSIZE classes and compute
	   the image's mean gray level mean1 */

	cnt1=0; mean1=0.0;
	for (y=0; y<img->ymax; y++)
		for (x=0; x<img->xmax; x++)
		{
			buf = readbuf_flt (img,x,y,zslice);
			bin = (int) (INT_HISTSIZE*(buf-min)/(max-min));
			if (bin>INT_HISTSIZE-1) bin=INT_HISTSIZE-1; else if (bin<0) bin=0;
			histy[bin] += 1.0;
			mean1+=buf; cnt1++;
		}
	mean1 /= cnt1;
	
	for (i=0; i<INT_HISTSIZE; i++) { histy[i] /= cnt1; histx[i]=histy[i]; }

	/* Run multiple passes of a 3-tap Gaussian filter over the histogram */

	for (j=0; j<10; j++)
	{
		histz[0] = 0.5*(histx[0]+histx[1]);
		histz[INT_HISTSIZE-1] = 0.5*(histx[INT_HISTSIZE-2]+histx[INT_HISTSIZE-1]);
		for (i=1; i<INT_HISTSIZE-1; i++) histz[i] = 0.25*(histx[i-1]+histx[i+1])+0.5*histx[i];
		for (i=0; i<INT_HISTSIZE; i++) histx[i] = histz[i];
	}

	/* To facilitate the next part, compute the central finite difference */

	histz[0] = 2*(histx[0]-histx[1]); histz[INT_HISTSIZE-1] = 2*(histx[INT_HISTSIZE-2]-histx[INT_HISTSIZE-1]);
	for (i=1; i<INT_HISTSIZE-1; i++) histz[i] = histx[i-1]-histx[i+1];

	/* Run over the first derivative. If the slope is initially positive, ignore.
		Start looking (status=1) when the slope turns negative, then stop (status=2)
		when the slope drops to 0 or positive. */

	status=0; i=0;
	do
	{
		if ((status==0) && (histz[i]>0) && (histz[i]>histz[i+1]) && (histz[i]>histz[i+2])) 
		{
			peak = histz[i]; 
			status=1;
		}
		else if ((status==1) && (histz[i]<0.01*peak)) status=2;
		if (status<2) i++; else thresh = min + i*(float)(max-min)/(float)INT_HISTSIZE;
	}
	while (( status<2) && (i<INT_HISTSIZE-1));

	if (mkgraph && !macrorun) 
	{
		for (i=0; i<INT_HISTSIZE; i++)
		{									/* Set 1: The first derivative */
			histz[INT_HISTSIZE+i] = histy[i];		/* Set 2: The original, unfiltered histogram */
			histz[512+i] = histx[i];		/* Set 3: The filtered histogram */
			histx[i] = min + i*(max-min)/INT_HISTSIZE;
		}
		multi_graphme (histx,histz,INT_HISTSIZE,3);
	}

	dp (1, "First valley threshold: %f\n", thresh);
	return thresh;


}






/* part 2 -- perform a hard thresholding on the image */

/* Note that we need to disable the statusbar if this is called
   from within any other (non-GUI) function */
   
/* NOTE: We determine threshold slice-by-slice. Is that good? */
/* TODO: Needs to handle RGB separately */

void do_thresholding_universal (image_type* img, int zslice, int what, int interactive)
{
unsigned char nullbuf[4];
int x,y,z, zl,zr;
int bx1,bx2,by1,by2;
float thresh;
char s[50];

	if (interactive && (img==&mainimg)) undo_nexus (img);

	memset (nullbuf, 0, sizeof (nullbuf));
	roi_bbox (img, &bx1, &by1, &bx2, &by2);	/* Observe ROI on main image */

	if (zslice >=0) 				/* Only single slice - compatibility mode, should be obsoleted */
	{
		zr=zl=zslice;
	}
	else
	{
		zl=0; zr=img->zmax-1;
	}
	
	for (z=zl; z<=zr; z++)
	{
		switch (what)
		{
			case 0:				/* Iterative */
				thresh = threshold_iterative (img, z);
				sprintf (s, "Iterative threshold value: %7.2f",thresh);
				break;
			case 1:
				thresh = otsus_threshold (img, z,1);
				sprintf (s, "Otsu's threshold value: %7.2f",thresh);
				break;
			case 2:
				thresh = entropy_threshold (img, z,1);
				sprintf (s, "Entropy threshold value: %7.2f",thresh);
				break;
			case 3:
				thresh = fvalley_threshold (img, z,1);
				sprintf (s, "First valley found at: %7.2f",thresh);
				break;
			default:
				thresh = 0.0;
		}
		if (interactive) statusbar_display (s);
		
		for (y=by1; y<=by2; y++)
			for (x=bx1; x<=bx2; x++)
			{
				if (in_roi(x,y))
				{
			
					if (readbuf_flt (img,x,y,z) < thresh)
						writebuf (img, nullbuf, x,y,z);
				}
			}
	}

	img->uid = next_uid++;

}



/******************************************************

	The opposite of thresholding: Clamp
	Clamp forces all image values I>T to T and
	I<-T to -T
	
*******************************************************/


void clamp_image (image_type *img, float thresh)
{
int x,y,z;
long idx;
int i; float buf;
int ithresh;
rgbtriplet rgb;


	ithresh = IROUND(thresh);
	if (img->imgtype == B_FLOAT)
		dp (1, "Clamping image values to range +/- %7.2f\n", thresh);
	else
		dp (1, "Clamping image values to range +/- %d\n", ithresh);


	for (z=0; z<img->zmax; z++)
		for (y=0; y<img->ymax; y++)
		{
			idx = img->xmax*(y+img->ymax*z);
			for (x=0; x<img->xmax; x++)
			{
				switch (img->imgtype)
				{
					case B_CHAR:
						i = ((unsigned char*)img->data)[idx];
						if (i>ithresh) 
							((unsigned char*)img->data)[idx] = ithresh;
						break;
					case B_SHORT:
						i = ((short*)img->data)[idx];
						if (i>ithresh) 
							((short*)img->data)[idx] = ithresh;
						else if (i<-ithresh) 
							((short*)img->data)[idx] = -ithresh;
						break;
					case B_FLOAT:
						buf = ((float*)img->data)[idx];
						if (buf>thresh) 
							((float*)img->data)[idx] = thresh;
						else if (buf<-thresh) 
							((float*)img->data)[idx] = -thresh;
						break;
					case B_RGB:
						rgb = ((rgbtriplet*)img->data)[idx];
						if (rgb.red>ithresh) rgb.red=ithresh;
						if (rgb.green>ithresh) rgb.green=ithresh;
						if (rgb.blue>ithresh) rgb.blue=ithresh;
						((rgbtriplet*)img->data)[idx] = rgb;
						break;

				}
				idx++;
			}
		}

	make_minmax (img);
	img->uid = next_uid++;

}



/***************************************************************************************

  ###    #   #   #####   #####   ####      #      ###    #####    ###    #   #   #####  
   #     #   #     #     #       #   #    # #    #   #     #       #     #   #   #      
   #     ##  #     #     #       #   #   #   #   #         #       #     #   #   #      
   #     # # #     #     ####    ####    #   #   #         #       #      # #    ####   
   #     #  ##     #     #       # #     #####   #         #       #      # #    #      
   #     #   #     #     #       #  #    #   #   #   #     #       #       #     #      
  ###    #   #     #     #####   #   #   #   #    ###      #      ###      #     #####  



	A standalone threshold function which is capable
	of handling both lower and upper limits --
	includes real-time preview, but always binarizes

	IS PRESENTLY DEACTIVATED AND AWAITING DEBUGGING
	
*******************************************************/

GtkAdjustment *threshold_minscale_adj;
GtkAdjustment *threshold_maxscale_adj;
int thresh_preview_type;


/* Simply reset to default colormap */

void threshold_reset_lut ()
{
int i;

	make_imlib_buffer_from_main_image (cur_slice);
	redraw (0);

}


void thresh_apply_lut (float l, float u, unsigned char preview)
{
int i,j, sidx, didx, x,y;
float min, max, f;
int imin,imax, xmax, ymax, stride;
GdkPixbuf *pixbuf;
unsigned char *im_buf;
rgbtriplet rgb;


	if (thresh_return.img->pixbuf)
	{

		/* Let us assume that xmax and thresh_return.img->xmax (and the corresponding ymax)
			match - careful when implementing a zoom operation */

		xmax = gdk_pixbuf_get_width (thresh_return.img->pixbuf);
		ymax = gdk_pixbuf_get_height (thresh_return.img->pixbuf);
		stride = gdk_pixbuf_get_rowstride (thresh_return.img->pixbuf);
		im_buf = gdk_pixbuf_get_pixels (thresh_return.img->pixbuf);			/* Pointer to pixbuf's rgb data */
		imgminmax (thresh_return.img, &min, &max);

		/* Distinguish between RGB and grayscale images. First, RGB images.
			A color preview is not allowed, and we always do a binary preview */

		if (thresh_return.img->imgtype==B_RGB)
		{

			for (y=0; y<ymax; y++)		/* we could use rgb_readbuf for additional safeguards, but... */
			{
				sidx = (y+cur_slice*thresh_return.img->ymax)*thresh_return.img->xmax;
				didx = y*stride;
				for (x=0; x<xmax; x++)
				{
					rgb = ((rgbtriplet *)thresh_return.img->data)[sidx++];
					if ((rgb.red  >=l) && (rgb.red  <=u)) im_buf[didx++]=255; else im_buf[didx++]=0;
					if ((rgb.green>=l) && (rgb.green<=u)) im_buf[didx++]=255; else im_buf[didx++]=0;
					if ((rgb.blue >=l) && (rgb.blue <=u)) im_buf[didx++]=255; else im_buf[didx++]=0;

				}
			}
		}
		else					/* Grayscale image, use readbuf_flt for this */
		{
			for (y=0; y<ymax; y++)
			{
				didx = y*stride;
				for (x=0; x<xmax; x++)
				{
					f = readbuf_flt (thresh_return.img, x, y, cur_slice);

					if (preview)				/* Color preview */
					{
						rgb.blue = (f-min)*255/(max-min);						/* Contrast-expanded gray */
						rgb.red = rgb.green = rgb.blue;
						if ((f>=l) && (f<=u))
						{
							rgb.red=255; rgb.green=0; rgb.blue=0;				/* Within threshold - bright red */
						}
					}
					else					/* Binary perview */
					{
						rgb = (rgbtriplet){0,0,0};
						if ((f>=l) && (f<=u))
						{
							rgb.red=255; rgb.green=255; rgb.blue=255;				/* Within threshold - white */
						}
					}
					im_buf[didx++] = rgb.red; im_buf[didx++] = rgb.green; im_buf[didx++] = rgb.blue; 
				}
			}
		}
		
		redraw (0);
	}
}


void on_threshold_minscale_adj_changed (GtkWidget *adj, gpointer user_data)
{
float f;

	f = threshold_minscale_adj->value;
	if (f > threshold_maxscale_adj->value) 
	{
		gtk_adjustment_set_value (threshold_maxscale_adj,f);
	}
	thresh_return.t = f;
	thresh_apply_lut (thresh_return.t, thresh_return.u, thresh_preview_type);

}

void on_threshold_maxscale_adj_changed (GtkWidget *adj, gpointer user_data)
{
float f;

	f = threshold_maxscale_adj->value;
	if (f < threshold_minscale_adj->value) 
	{
		gtk_adjustment_set_value (threshold_minscale_adj,f);
	}
	thresh_return.u = f;
	thresh_apply_lut (thresh_return.t, thresh_return.u, thresh_preview_type);

}


void on_pv_menu_changed (GtkWidget *widget, gpointer user_data)
{
int i;

	i = *(int*) user_data;
	thresh_preview_type=i;
	thresh_apply_lut (thresh_return.t, thresh_return.u, thresh_preview_type);

}


void on_threshold_ok_clicked ( GtkButton * button, gpointer data )
{
int i;

	undo_nexus (&mainimg);
	threshold_reset_lut ();

	do_thresholding2 (thresh_return.t, thresh_return.u, -1);
	
	gtk_widget_destroy ( GTK_WIDGET ( data ) );
}


void on_threshold_cancel_clicked ( GtkButton * button, gpointer data )
{
int i;


	threshold_reset_lut ();
	statusbar_display (" ");

	gtk_widget_destroy ( GTK_WIDGET ( data ) );
}


void create_thresh_wnd (image_type* img)
{
  GtkWidget *thresh_wnd;
  GtkWidget *threshold_vbox1;
  GtkWidget *threshold_minscale;
  GtkWidget *threshold_maxscale;
  GtkWidget *threshold_hbox1;
  GtkWidget *threshold_okbutton;
  GtkWidget *threshold_cancelbutton;
  GtkWidget *pv_menu, *pv_menu_menu;
  GtkWidget *menuitem;
  GtkWidget *hsep;

  float min, max, stepsz;
  

	imgminmax (img, &min, &max);
	thresh_return.t = min; thresh_return.u = max;
	thresh_return.thresh_type = TH_BINARY;
	stepsz=1.0; 	/* for all int images (rgb, char, short) */
	if (img->imgtype==B_FLOAT) stepsz=0.1;
	thresh_preview_type=0;
	
	statusbar_display ("Previewing threshold");

	thresh_wnd = gtk_window_new (GTK_WINDOW_TOPLEVEL);
	gtk_object_set_data (GTK_OBJECT (thresh_wnd), "thresh_wnd", thresh_wnd);
	gtk_window_set_title (GTK_WINDOW (thresh_wnd), "Dual Threshold");
	gtk_window_set_modal (GTK_WINDOW (thresh_wnd), TRUE);
	gtk_window_set_default_size (GTK_WINDOW(thresh_wnd), 220,150);		/* Sliders should have a minimum length */

	threshold_vbox1 = vbox_in_window (thresh_wnd,"threshold_vbox1");

	threshold_minscale_adj = GTK_ADJUSTMENT (gtk_adjustment_new (min, min, max, stepsz, 10*stepsz, 0));
	threshold_minscale = MakeHScale (GTK_ADJUSTMENT(threshold_minscale_adj), thresh_wnd, threshold_vbox1, "threshold_minscale", -1,-1);
	gtk_signal_connect (GTK_OBJECT (threshold_minscale_adj), "value_changed", GTK_SIGNAL_FUNC (on_threshold_minscale_adj_changed),threshold_minscale_adj);

	threshold_maxscale_adj = GTK_ADJUSTMENT (gtk_adjustment_new (max, min, max, stepsz, 10*stepsz, 0));
	threshold_maxscale = MakeHScale (GTK_ADJUSTMENT(threshold_maxscale_adj), thresh_wnd, threshold_vbox1, "threshold_maxscale", -1,-1);
	gtk_signal_connect (GTK_OBJECT (threshold_maxscale_adj), "value_changed", GTK_SIGNAL_FUNC (on_threshold_maxscale_adj_changed),threshold_maxscale_adj);

	if (img->imgtype!=B_FLOAT)
	{
		gtk_scale_set_digits (GTK_SCALE (threshold_minscale), 0);
		gtk_scale_set_digits (GTK_SCALE (threshold_maxscale), 0);
	}

	pv_menu = Make_Optionmenu (thresh_wnd,threshold_vbox1,"pv_menu",-1,-1);
	pv_menu_menu = gtk_menu_new ();
	menuitem = Option_menuitem (pv_menu_menu, GTK_SIGNAL_FUNC(on_pv_menu_changed), "Binary preview", (gpointer) &menuopt0);
	menuitem = Option_menuitem (pv_menu_menu, GTK_SIGNAL_FUNC(on_pv_menu_changed), "Color preview", (gpointer) &menuopt1);
	gtk_option_menu_set_menu (GTK_OPTION_MENU (pv_menu), pv_menu_menu);

	hsep = MakeHsep (thresh_wnd, threshold_vbox1, "hseparator4", -1,-1);
	threshold_hbox1 = hbox_in_window (thresh_wnd, threshold_vbox1, NULL);

	threshold_okbutton = Button_in_table (thresh_wnd,threshold_hbox1,"OK","threshold_okbutton",-1,-1,NULL);
	threshold_cancelbutton = Button_in_table (thresh_wnd,threshold_hbox1,"Cancel","threshold_cancelbutton",-1,-1,NULL);

	gtk_signal_connect (GTK_OBJECT (thresh_wnd), "destroy",
                      GTK_SIGNAL_FUNC (on_threshold_cancel_clicked), thresh_wnd);
	gtk_signal_connect (GTK_OBJECT (threshold_okbutton), "clicked",
                      GTK_SIGNAL_FUNC (on_threshold_ok_clicked),thresh_wnd);
	gtk_signal_connect (GTK_OBJECT (threshold_cancelbutton), "clicked",
                      GTK_SIGNAL_FUNC (on_threshold_cancel_clicked),thresh_wnd);

	gtk_widget_show (thresh_wnd);
}


void dual_threshold (image_type* img)
{

	thresh_return.img = img;
	create_thresh_wnd (img);
}





/*********************************************************************

   #     ####      #     ####    #####    ###    #   #   #####  
  # #    #   #    # #    #   #     #       #     #   #   #      
 #   #   #   #   #   #   #   #     #       #     #   #   #      
 #   #   #   #   #   #   ####      #       #      # #    ####   
 #####   #   #   #####   #         #       #      # #    #      
 #   #   #   #   #   #   #         #       #       #     #      
 #   #   ####    #   #   #         #      ###      #     #####  

	Adaptive (local) threshold functions

************************************************************************/

/* NOTE - these functions are VERY slow. Somehow we should speed them up. */

/* Here is a strategy to speed 'em up. Have another image or matrix of values. 
	This matrix is M x M pixels, where M is the local window size. From one column 
	to the next, we simply delete the leftmost column and add the new rightmost one.
	This trick works for local mean (and we could do some housekeeping for Otsu),
	but it wouldn't work for Isodata. */
	


/* Local threshold by local mean value */

float get_local_threshold_m (image_type *img, int x, int y, int z, int radius, int oorflag)
{
int x1,y1;
float mean, buf;
int cnt;


	cnt=0; mean=0.0; 
	for (y1=y-radius; y1<=y+radius; y1++)
	{
		for (x1=x-radius; x1<=x+radius; x1++)
		{
			if ((y1>=0) && (x1>=0) && (x1<img->xmax) && (y1<img->ymax))
			{
				buf = readbuf_flt (img,x1,y1,z);
				mean += buf; cnt++;
			}
		}
	}

	if (cnt>0) 
		return mean / cnt;
	else 
		return 0;
}



/* Local threshold by iterative (isodata) threshold finding */

float get_local_threshold_i (image_type *img, int x, int y, int z, int radius, int oorflag)
{
int x1,y1;
float min, max;
float thresh,newthresh,buf,mean1,mean2, delta;
int cnt1,cnt2;
int abort, pass;


	imgminmax (img, &min, &max);
	abort=0; pass=0;
	
	/* Initial threshold collects everything in mean1 */

	thresh = min-1.0;
	delta = max-min;

	/* Start iterations */
	
	do
	{
		cnt1=0; cnt2=0; mean1=0.0; mean2=0.0;
		for (y1=y-radius; y1<=y+radius; y1++)
		{
			for (x1=x-radius; x1<=x+radius; x1++)
			{
				if ((y1>=0) && (x1>=0) && (x1<img->xmax) && (y1<img->ymax))
				{
					buf = readbuf_flt (img,x1,y1,z);
					if (buf > thresh)
					{
						mean1 += buf; cnt1++;
					}
					else
					{
						mean2 += buf; cnt2++;
					}
				}
			}
		}

		if (cnt1>0) mean1 /= cnt1;
		if (cnt2>0) mean2 /= cnt2;
		
		if ((cnt1==0) && (cnt2==0))
			{ abort=1; newthresh=0; }		/* bad error */
		else if (cnt1==0)
			newthresh = mean2;
		else if (cnt2==0)
			newthresh = mean1;
		else
			newthresh = 0.5*(mean1+mean2);
		
		if ( fabs(thresh - newthresh) < 0.001*delta )
			abort=1;
		pass++;
		thresh = newthresh;
		dp (3, "Pass %d: Mean 1 %f (area %d), mean 2 %f (area %d), new threshold %f\n",
			pass, mean1, cnt1, mean2, cnt2, newthresh);
	}
	while (!abort);
	
	return thresh;
}


/* Distributor for the individual threshold finding methods. Note that
	the threshold finders do NOT observe the ROI */

float get_local_threshold (image_type *img, int x, int y, int z, int radius, int oorflg, int method)
{
	if (method==0)
		return get_local_threshold_m (img, x,y,z, radius, oorflg);
	else if (method==1)
		return get_local_threshold_i (img, x,y,z, radius, oorflg);
}



void adaptive_threshold (image_type* img, int radius, float bias, int oorflg, int method)
{
int x,y,x1,x2,y1,y2, z, i;
int bx1, by1, bx2, by2;
long idx,idx2;
image_type img2, rgbimg;
rgbtriplet rgb;
float T;
unsigned char nullelem[4] = {0, 0, 0, 0};
	
	
	/* Clone one slice of the the current image to hold the 
		threshold map. Note that we need to store a full threshold map
		before the actual thresholding, because we'd otherwise bias
		the local threshold */
	
	if (allocate_image (&img2, img->imgtype, img->xmax, img->ymax, 1) == -1)
		fatal_memerr();	

	if (img->imgtype == B_RGB)
	{
		if (allocate_image (&rgbimg, B_CHAR, img->xmax, img->ymax, 1) == -1)
			fatal_memerr();	
	}
	
	roi_bbox (img, &bx1, &by1, &bx2, &by2);	/* Observe ROI */

	for (z=0; z<img->zmax; z++)
	{

		if (img->imgtype == B_RGB)
		{
			for (i=0; i<3; i++)				/* Handle R, G, B separately (slow!!) */
			{
				for (y=0; y<img->ymax; y++)
					for (x=0; x<img->xmax; x++)
					{
						rgb = readbuf_rgb (img, x,y,z);
						((unsigned char*)rgbimg.data)[x + img->xmax*y] = 
								(i==0? rgb.red : (i==1? rgb.green : rgb.blue));
					}

				for (y=by1; y<=by2; y++)
				{
					if (!macrorun)
						indicate_progress ( (float)((i+1)*(z+1)*(y-by1))/(float)(3*img->zmax*(by2-by1)));
				
					for (x=bx1; x<=bx2; x++)
					{
						if (in_roi (x,y))
						{
							T = get_local_threshold (&rgbimg, x,y,z, radius, oorflg, method);
							idx = x + img2.xmax*y;
							((unsigned char*)img2.data)[idx] = CLAMP ((int)(T+bias+0.499), 0, 255);
						}
					}
				}

				for (y=by1; y<=by2; y++)
					for (x=bx1; x<=bx2; x++)
					{
						if (in_roi (x,y))
						{
							rgb = readbuf_rgb (img, x,y,z);
							if (i==0)
							{
								if (rgb.red < ((unsigned char*)img2.data)[x+img2.xmax*y]) rgb.red=0;
							}
							else if (i==1)
							{
								if (rgb.green < ((unsigned char*)img2.data)[x+img2.xmax*y]) rgb.green=0;
							}
							else if (i==2)
							{
								if (rgb.blue < ((unsigned char*)img2.data)[x+img2.xmax*y]) rgb.blue=0;
							}
							writebuf (img, &rgb, x,y,z);		/* Hard thresholding */
						}
					}

			}
		}
		else					/* All non-RGB types */
		{
			for (y=by1; y<=by2; y++)
			{
				if (!macrorun)
					indicate_progress ( (float)((z+1)*(y-by1))/(float)(img->zmax*(by2-by1)));
			
				for (x=bx1; x<=bx2; x++)
				{
					if (in_roi (x,y))
					{
						T = get_local_threshold (img, x,y,z, radius, oorflg, method);
						idx = x + img2.xmax*y;
						switch (img2.imgtype)
						{
							case B_CHAR:
								((unsigned char*)img2.data)[idx] = CLAMP((int)(T+bias+0.499), 0,255);
								break;
							case B_SHORT:
								((short*)img2.data)[idx] = (int)(T+bias+0.499);
								break;
							case B_FLOAT:
								((float*)img2.data)[idx] = T+bias;
								break;
						}

					}

				}
			}
		}

		/* One slice completed. Perform the actual thresholding */
		
		for (y=by1; y<=by2; y++)
		{
			for (x=bx1; x<=bx2; x++)
			{
				if (in_roi (x,y))
				{
						if (readbuf_flt (img,x,y,z) < readbuf_flt (&img2,x,y,0) )
							writebuf (img, nullelem, x,y,z);		/* Hard thresholding */
				}
			}
		}
	}


	if (!macrorun)
	{
		make_minmax (&img2);
		secondary_display (&img2, 0, "Threshold map");
		reset_progress();
	}


	freebuf (&img2);
	make_minmax (img);
	img->uid = next_uid++;
}
		





GtkAdjustment *athreshold_winsize_adj;
GtkAdjustment *athreshold_bias_adj;
int athreshold_method;

/* TODO: There is another problem that comes up with the progressbar.
	The window now receives signals and can actually enter the 
	processing function recursively. Not sure how to handle this, yet */


void on_athreshold_ok_clicked ( GtkButton * button, gpointer data )
{
int radius;
float bias;

	radius = athreshold_winsize_adj->value;
	bias = athreshold_bias_adj->value;

	undo_nexus (&mainimg);
	adaptive_threshold (&mainimg, radius, bias, 0, athreshold_method);
	make_imlib_buffer_from_main_image (cur_slice); 
	redraw (0);

	gtk_widget_destroy ( GTK_WIDGET ( data ) );
}


void on_athreshold_cancel_clicked ( GtkButton * button, gpointer data )
{
int i;

	gtk_widget_destroy ( GTK_WIDGET ( data ) );
}



void on_at_menu_changed (GtkWidget *widget, gpointer user_data)
{
int i;

	i = *(int*) user_data;
	athreshold_method = i;
}






void create_athresh_wnd (image_type* img)
{
  GtkWidget *athresh_wnd;
  GtkWidget *athreshold_vbox1;
  GtkWidget *athreshold_winsize;
  GtkWidget *athreshold_bias;
  GtkWidget *athreshold_hbox1;
  GtkWidget *athreshold_okbutton;
  GtkWidget *athreshold_cancelbutton;
  GtkWidget *at_menu, *at_menu_menu;
  GtkWidget *menuitem;
  GtkWidget *hsep, *table1, *label;
  GtkTooltips *tt_method, *tt_window, *tt_bias, *tt_okbtn, *tt_canbtn;
  
  float min, max, stepsz;
  int y;
  

	imgminmax (img, &min, &max);
	stepsz=1.0; 	/* for all int images (rgb, char, short) */
	if (img->imgtype==B_FLOAT) stepsz=0.1;
	athreshold_method = 0;

	athresh_wnd = gtk_window_new (GTK_WINDOW_TOPLEVEL);
	gtk_object_set_data (GTK_OBJECT (athresh_wnd), "athresh_wnd", athresh_wnd);
	gtk_window_set_title (GTK_WINDOW (athresh_wnd), "Adaptive Thresholding");
	gtk_window_set_modal (GTK_WINDOW (athresh_wnd), TRUE);
	gtk_window_set_default_size (GTK_WINDOW(athresh_wnd), 220,150);		/* Sliders should have a minimum length */

	athreshold_vbox1 = vbox_in_window (athresh_wnd,"athreshold_vbox1");
	table1 = MakeTable (athresh_wnd, athreshold_vbox1, "athresh_table1", 2, 3);

	y=0;
	label = MakeLabel (athresh_wnd,table1,"Operation", NULL,0,y);
	at_menu = Make_Optionmenu (athresh_wnd,table1,"at_menu",1,y);
	at_menu_menu = gtk_menu_new ();
	menuitem = Option_menuitem (at_menu_menu, GTK_SIGNAL_FUNC(on_at_menu_changed), "Local mean", (gpointer) &menuopt0);
	menuitem = Option_menuitem (at_menu_menu, GTK_SIGNAL_FUNC(on_at_menu_changed), "Iterative", (gpointer) &menuopt1);
	gtk_option_menu_set_menu (GTK_OPTION_MENU (at_menu), at_menu_menu);
	tt_method = gtk_tooltips_new ();
	gtk_tooltips_set_tip (tt_method, at_menu, "Choose the method by which a local window decides whether its\npixel is reset to 0 or kept unchanged", NULL);

	y++;
	label = MakeLabel (athresh_wnd,table1,"Window radius", NULL,0,y);
	athreshold_winsize_adj = GTK_ADJUSTMENT (gtk_adjustment_new (40, 5, 256, 1, 10, 0));
	athreshold_winsize = MakeHScale (GTK_ADJUSTMENT(athreshold_winsize_adj), athresh_wnd, table1, "athreshold_winsize", 1,y);
	tt_window = gtk_tooltips_new ();
	gtk_tooltips_set_tip (tt_window, athreshold_winsize, "Radius of the local window.\nThe local window is a rectangle with a side length of twice the radius", NULL);
	//gtk_signal_connect (GTK_OBJECT (athreshold_minscale_adj), "value_changed", GTK_SIGNAL_FUNC (on_athreshold_minscale_adj_changed),athreshold_minscale_adj);

	y++;
	label = MakeLabel (athresh_wnd,table1,"Bias", NULL,0,y);
	athreshold_bias_adj = GTK_ADJUSTMENT (gtk_adjustment_new (0, -max, max, stepsz, 10*stepsz, 0));
	athreshold_bias = MakeHScale (GTK_ADJUSTMENT(athreshold_bias_adj), athresh_wnd, table1, "athreshold_bias", 1,y);
	tt_bias = gtk_tooltips_new ();
	gtk_tooltips_set_tip (tt_bias, athreshold_bias, "This value is added to the local threshold.\nHigher value means higher effective threshold", NULL);

	//gtk_signal_connect (GTK_OBJECT (athreshold_maxscale_adj), "value_changed", GTK_SIGNAL_FUNC (on_athreshold_maxscale_adj_changed),athreshold_maxscale_adj);

	if (img->imgtype!=B_FLOAT)
	{
		gtk_scale_set_digits (GTK_SCALE (athreshold_bias), 0);
	}

	hsep = MakeHsep (athresh_wnd, athreshold_vbox1, "hseparator4", -1,-1);

	athreshold_hbox1 = hbox_in_window (athresh_wnd, athreshold_vbox1, NULL);

	athreshold_okbutton = Button_in_table (athresh_wnd,athreshold_hbox1,"OK","athreshold_okbutton",-1,-1,NULL);
	athreshold_cancelbutton = Button_in_table (athresh_wnd,athreshold_hbox1,"Cancel","athreshold_cancelbutton",-1,-1,NULL);
	tt_okbtn = gtk_tooltips_new ();
	tt_canbtn = gtk_tooltips_new ();
	gtk_tooltips_set_tip (tt_okbtn, athreshold_okbutton, "Hit OK to start the process\nNOTE - THIS FUNCTION IS VERY SLOW", NULL);
	gtk_tooltips_set_tip (tt_canbtn, athreshold_cancelbutton, "Hit CANCEL to back out", NULL);

	gtk_signal_connect (GTK_OBJECT (athresh_wnd), "destroy",
                      GTK_SIGNAL_FUNC (on_athreshold_cancel_clicked), athresh_wnd);
	gtk_signal_connect (GTK_OBJECT (athreshold_okbutton), "clicked",
                      GTK_SIGNAL_FUNC (on_athreshold_ok_clicked),athresh_wnd);
	gtk_signal_connect (GTK_OBJECT (athreshold_cancelbutton), "clicked",
                      GTK_SIGNAL_FUNC (on_athreshold_cancel_clicked),athresh_wnd);

	gtk_window_set_modal (GTK_WINDOW(athresh_wnd), TRUE);
	gtk_widget_show (athresh_wnd);
}


